# -*- coding: utf-8 -*-
"""
Twitter's OAuth Demonstratin on Google App Engine

License: The MIT license
Copyright (c) 2009 furyu-tei
"""

import os,cgi,re,urllib,urlparse
import logging
import wsgiref.handlers
import time,datetime
import random,hashlib,hmac,base64

from google.appengine.ext import db
from google.appengine.ext import webapp
from google.appengine.ext.webapp import template

from gae_twitoauth import GAE_TwitOAuth

BASE_PATH = u'/demo'

DEBUG_FLAG = True           # for webapp.WSGIApplication()
DEBUG_LEVEL = logging.DEBUG # for logger
DEBUG=True                  # False:  disable log() (wrapper of logging.debug())

DEMO_REQUEST_LIMIT_HOURS = 1
DB_FETCH_LIMIT = 100 # max fetch number of datastore

class dbOAuthDemoRequest(db.Expando):
  req_token = db.StringProperty()
  date = db.DateTimeProperty(auto_now_add=True)


class dbOAuthDemoUser(db.Expando):
  uname = db.StringProperty()
  screen_name = db.StringProperty()
  auth_key = db.StringProperty()
  form_token = db.StringProperty()
  last_logout = db.DateTimeProperty()
  date = db.DateTimeProperty(auto_now_add=True)


def log(*args):
  if not DEBUG: return
  for arg in args:
    try:
      logging.debug(arg)
    except Exception, s:
      logging.error(u'*** Error in log(): %s' % (unicode(s)))
      pass


def logerr(*args):
  for arg in args:
    try:
      logging.error(arg)
    except Exception, s:
      logging.error(u'*** Error in logerr(): %s' % (unicode(s)))
      pass


def cleanup_DemoRequest():
  threshold = datetime.datetime.utcnow()-datetime.timedelta(hours=DEMO_REQUEST_LIMIT_HOURS)
  db_demo_req = dbOAuthDemoRequest.all().filter('date <= ',threshold)
  while True:
    db_req_list = db_demo_req.fetch(DB_FETCH_LIMIT)
    len_req = len(db_req_list)
    if len_req <=0:
      break
    try:
      db.delete(db_req_list)
    except Exception, s:
      logerr(u'cleanup_DemoRequest(): %s' % (s))
      break
    if len_req<DB_FETCH_LIMIT:
      break


def get_hash_rand_hexdigest(pre_str=None):
  if not pre_str: pre_str = u''
  ustr = u'%s-%16x' % (pre_str,random.getrandbits(64))
  hexdigest = hashlib.sha224(ustr.encode('utf-8')).hexdigest()
  return hexdigest


def get_demo_request(req_id=None,req_token=None,flg_create=False):
  if req_id:
    _key = db.Key.from_path('dbOAuthDemoRequest',int(req_id))
    log(_key)
    db_req = db.get(_key)
  elif req_token:
    db_req = dbOAuthDemoRequest.all().order('-date').filter('req_token = ',req_token).get()
  elif flg_create:
    req_token = get_hash_rand_hexdigest()
    db_req = dbOAuthDemoRequest(req_token=req_token)
    db_req.put()
  else:
    db_req = None
  return db_req


def get_demo_user(user_id=None,screen_name=None,uname=None,flg_create=False):
  db_user = None
  while True:
    if user_id:
      user_id = unicode(user_id)
      _key = db.Key.from_path('dbOAuthDemoUser',user_id)
      db_user = db.get(_key)
      if not flg_create or not screen_name:
        break
      
      if not uname:
         uname = screen_name
      if db_user: # update
        if db_user.screen_name!= screen_name or db_user.uname!=uname:
          form_token = get_hash_rand_hexdigest(screen_name)
          db_user.uname = uname
          db_user.screen_name = screen_name
          db_user.form_token = form_token
          db_user.put()
        break
      
      def _create(user_id,screen_name):
        auth_key = get_hash_rand_hexdigest(user_id)
        form_token = get_hash_rand_hexdigest(screen_name)
        db_user = dbOAuthDemoUser(
          key_name = user_id,
          uname = uname,
          screen_name = screen_name,
          auth_key = auth_key,
          form_token = form_token,
        )
        db_user.put()
        return db_user
      try:
        db_user = db.run_in_transaction(_create,user_id,screen_name)
      except Exception, s:
        logerr(u'Error in get_demo_user(): %s' % (s))
        db_user = db.get(_key)
      break
    
    if screen_name:
      db_user = dbOAuthDemoUser.all().order('-date').filter('screen_name = ').get()
      break
    
    break
  
  return db_user


def createCookieStr(name,val,req,domain=None,path=None,limitdays=366):
  curtime=datetime.datetime.utcnow()
  umems=urlparse.urlparse(req.url)
  cvals=[]
  cvals.append('%s=%s' % (name,urllib.quote(val.encode('utf-8'))))
  if domain:
    cvals.append('domain=%s' % (urllib.quote(domain.encode('utf-8'))))
  else:
    if not re.search(r'^localhost(:\d+)?$',umems.netloc):
      cvals.append('domain=%s' % (urllib.quote(re.sub(r':.*$','',umems.netloc).encode('utf-8'))))
  if path:
    cvals.append('path=%s' % (urllib.quote(path.encode('utf-8'))))
  else:
    cvals.append('path=%s' % (urllib.quote(umems.path.encode('utf-8'))))
  cvals.append('expires='+(curtime+datetime.timedelta(days=limitdays)).strftime("%a, %d %b %Y %H:%M:%S GMT"))
  #cvals.append('httponly')
  
  return '; '.join(cvals)


def createRemoveCookieStr(name,req,domain=None,path=None):
  curtime = datetime.datetime.utcnow()
  umems = urlparse.urlparse(req.url)
  cvals=[]
  cvals.append('%s=' % (name))
  if domain:
    cvals.append('domain=%s' % (urllib.quote(domain.encode('utf-8'))))
  else:
    if not re.search(r'^localhost(:\d+)?$',umems.netloc):
      cvals.append('domain=%s' % (urllib.quote(re.sub(r':.*$','',umems.netloc).encode('utf-8'))))
  if path:
    cvals.append('path=%s' % (urllib.quote(path.encode('utf-8'))))
  else:
    cvals.append('path=%s' % (urllib.quote(umems.path.encode('utf-8'))))
  cvals.append('expires='+(curtime-datetime.timedelta(days=1)).strftime("%a, %d %b %Y %H:%M:%S GMT"))
  #cvals.append('httponly')
  
  return '; '.join(cvals)


def common_init(self):
  (req,rsp) = (self.request,self.response)
  (rheaders,rcookies) = (req.headers,req.cookies)
  self.currentpath=os.path.dirname(__file__)
  self.tplpath=os.path.join(self.currentpath,'template/')
  self.urlbase=re.sub(u'[?#].*$','',req.url)
  self.urlroot=re.sub(u'(http://[^/]*).*$',r'\1',req.url)
  return (req,rsp,rheaders,rcookies)


class demo(webapp.RequestHandler):
  def _common(self):
    (req,rsp,rheaders,rcookies)=common_init(self)
    
    user_id = rcookies.get('oauth_demo_user_id')
    db_user = get_demo_user(user_id=user_id)
    if db_user and db_user.auth_key == rcookies.get('oauth_demo_auth_key'):
      twitter = GAE_TwitOAuth(user_id=user_id)
    else:
      db_user = None
      twitter = GAE_TwitOAuth()
    
    tvalues={}
    if twitter.oauth_enable():
      logout_url = twitter.get_logout_url(callback_url=BASE_PATH+'/cb_logout')
      uname = db_user.uname
      #screen_name = twitter.get_screen_name()
      screen_name = db_user.screen_name
      form_token = db_user.form_token
      tvalues.update({
        'logout_url': logout_url,
        'uname': uname,
        'screen_name': screen_name,
        'form_token': form_token,
      })
      if db_user.last_logout:
        tvalues['last_logout'] = re.sub(u'-',r'/',re.sub(u'T',r'&nbsp;',re.sub(u'\.\d+$',r'',(db_user.last_logout+datetime.timedelta(hours=9)).isoformat())))
      
      if req.get('update_status')=='1':
        show_message = u'投稿に失敗しちゃいました…'
        if req.get('form_token')==form_token:
          status = req.get('status',u'')[:140]
          (json,result_code) = twitter.update_status(status)
          if 200<=result_code<300:
            tvalues['show_message'] = u'つぶやきを投稿しましたっ！'
            tvalues['permalink'] = u'http://twitter.com/%s/status/%s' % (screen_name,unicode(json['id']))
    else:
      tvalues['login_url'] = BASE_PATH + '/login'
    
    tfile=os.path.join(self.tplpath,'demo_twitoauth.html')
    rsp.headers['Content-Type']='text/html'
    rsp.out.write(template.render(tfile, tvalues))

  def get(self):
    self._common()
  
  def post(self):
    self._common()


class login(webapp.RequestHandler):
  def get(self):
    (req,rsp,rheaders,rcookies)=common_init(self)
    
    cleanup_DemoRequest()
    db_req = get_demo_request(flg_create=True)
    
    twitter = GAE_TwitOAuth()
    login_url = twitter.get_login_url(BASE_PATH+'/cb_login?req_token='+db_req.req_token)
    
    rsp.headers.add_header('Set-Cookie',createCookieStr('oauth_demo_req_id',str(db_req.key().id()),req,path=BASE_PATH))
    self.redirect(login_url)


class cb_login(webapp.RequestHandler):
  def get(self):
    (req,rsp,rheaders,rcookies)=common_init(self)
    
    flg_success = False
    while True:
      db_req = get_demo_request(req_id=rcookies.get('oauth_demo_req_id'))
      if not db_req:
        logerr('has no token')
        break
      
      if db_req.req_token != req.get('req_token'):
        logerr('unmatch request token')
        break
      
      db.delete(db_req)
      rsp.headers.add_header('Set-Cookie',createRemoveCookieStr('oauth_demo_req_id',req,path=BASE_PATH))
      
      user_id = req.get('user_id')
      screen_name = req.get('screen_name')
      if not user_id or not screen_name:
        log('unauthorized')
        break
      
      twitter = GAE_TwitOAuth(user_id=user_id)
      uname = None
      if twitter.oauth_enable():
        #(json,status_code)=twitter.show(id=user_id)
        #(json,status_code)=twitter.show(id=screen_name)
        (json,status_code)=twitter.show(user_id=user_id)
        #(json,status_code)=twitter.show(screen_name=screen_name)
        if 200<=status_code<300:
          uname = json.get('name',u'')
          screen_name = json.get('screen_name',u'')
      
      db_user = get_demo_user(user_id=user_id,screen_name=screen_name,uname=uname,flg_create=True)
      if not db_user:
        logerr('datestore failure')
        break
      
      rsp.headers.add_header('Set-Cookie',createCookieStr('oauth_demo_user_id',user_id,req,path=BASE_PATH))
      rsp.headers.add_header('Set-Cookie',createCookieStr('oauth_demo_auth_key',db_user.auth_key,req,path=BASE_PATH))
      rsp.headers.add_header('P3P','CP="CAO PSA OUR"')
      
      flg_success = True
      break
    
    self.redirect(BASE_PATH)


class cb_logout(webapp.RequestHandler):
  def get(self):
    (req,rsp,rheaders,rcookies)=common_init(self)
    
    user_id = rcookies.get('oauth_demo_user_id')
    db_user = get_demo_user(user_id=user_id)
    if db_user and db_user.auth_key == rcookies.get('oauth_demo_auth_key'):
      db_user.last_logout = datetime.datetime.utcnow()
      db_user.put()
    self.redirect(BASE_PATH)


def main():
  logging.getLogger().setLevel(DEBUG_LEVEL)
  application = webapp.WSGIApplication([
    (BASE_PATH+'/login', login),
    (BASE_PATH+'/cb_login', cb_login),
    (BASE_PATH+'/cb_logout', cb_logout),
    (BASE_PATH, demo),
  ],debug=DEBUG_FLAG)
  wsgiref.handlers.CGIHandler().run(application)

if __name__ == "__main__":
  main()

#■ end of file
